@@ -1,10 +1,7 @@
 from __future__ import (absolute_import, division, print_function,
                         unicode_literals)
-
 import six
 from six import with_metaclass
-from six.moves import range
-import os
 import numpy as np
 import itertools
 from slicerator import Slicerator, propagate_attr, index_attr
@@ -288,6 +285,152 @@ Pixel Datatype: {dtype}""".format(w=self.frame_shape[0],
                                   count=len(self),
                                   dtype=self.pixel_type)
 
+def _iter_attr(obj):
+    try:
+        for ns in [obj] + obj.__class__.mro():
+            for attr in ns.__dict__:
+                yield ns.__dict__[attr]
+    except AttributeError:
+        raise StopIteration  # obj has no __dict__
+
+
+def _transpose(get_frame, expected_axes, desired_axes):
+    if list(expected_axes) == list(desired_axes):
+        return get_frame
+    else:
+        transposition = [expected_axes.index(a) for a in desired_axes]
+        def get_frame_T(**ind):
+            return get_frame(**ind).transpose(transposition)
+        return get_frame_T
+
+
+def _bundle(get_frame, expected_axes, to_iter, sizes, dtype):
+    bundled_axes = to_iter + expected_axes
+    shape = [sizes[a] for a in bundled_axes]
+    iter_shape = shape[:len(to_iter)]
+    def get_frame_bundled(**ind):
+        result = np.empty(shape, dtype=dtype)
+        md_list = []
+        for indices in itertools.product(*[range(s) for s in iter_shape]):
+            ind.update({n: i for n, i in zip(to_iter, indices)})
+            frame = get_frame(**ind)
+            result[indices] = frame
+
+            if hasattr(frame, 'metadata'):
+                if frame.metadata is not None:
+                    md_list.append(frame.metadata)
+        # propagate metadata
+        if len(md_list) == np.prod(iter_shape):
+            metadata = dict()
+            keys = md_list[0].keys()
+            for k in keys:
+                try:
+                    metadata[k] = [row[k] for row in md_list]
+                except KeyError:
+                    # if a field is not present in every frame, ignore it
+                    warn('metadata field {} is not propagated')
+                else:
+                    # if all values are equal, only return one value
+                    if metadata[k][1:] == metadata[k][:-1]:
+                        metadata[k] = metadata[k][0]
+                    else:  # cast into ndarray
+                        metadata[k] = np.array(metadata[k])
+                        metadata[k].shape = iter_shape
+        else:
+            metadata = None
+        return Frame(result, metadata=metadata)
+    return get_frame_bundled, bundled_axes
+
+
+def _drop(get_frame, expected_axes, to_drop):
+    # sort axes in descending order for correct function of np.take
+    to_drop_inds = [list(expected_axes).index(a) for a in to_drop]
+    indices = np.argsort(to_drop_inds)
+    axes = [to_drop_inds[i] for i in reversed(indices)]
+    to_drop = [to_drop[i] for i in reversed(indices)]
+    result_axes = [a for a in expected_axes if a not in to_drop]
+
+    def get_frame_dropped(**ind):
+        result = get_frame(**ind)
+        for (ax, name) in zip(axes, to_drop):
+            result = np.take(result, ind[name], axis=ax)
+        return result
+    return get_frame_dropped, result_axes
+
+
+def _make_get_frame(result_axes, get_frame_dict, sizes, dtype):
+    methods = list(get_frame_dict.keys())
+    result_axes = [a for a in result_axes]
+    result_axes_set = set(result_axes)
+
+    # search for get_frame methods that return the right axes
+    for axes in methods:
+        if len(set(axes) ^ result_axes_set) == 0:
+            # _transpose does nothing when axes == result_axes
+            return _transpose(get_frame_dict[axes], axes, result_axes)
+
+    # we need either to drop axes or to iterate over axes:
+    # collect some numbers to decide what to do
+    arr = [None] * len(methods)
+    for i, method in enumerate(methods):
+        axes_set = set(method)
+        to_iter_set = result_axes_set - axes_set
+        to_iter = [x for x in result_axes if x in to_iter_set]  # fix the order
+        n_iter = int(np.prod([sizes[ax] for ax in to_iter]))
+        to_drop = list(axes_set - result_axes_set)
+        n_drop = int(np.prod([sizes[ax] for ax in to_drop]))
+        arr[i] = [method, axes_set, to_iter, n_iter, to_drop, n_drop]
+
+    # try to read as less data as possible: try n_drop == 0
+    # sort in increasing number of iterations
+    arr.sort(key=lambda x: x[3])
+    for method, axes_set, to_iter, n_iter, to_drop, n_drop in arr:
+        if n_drop > 0:
+            continue
+        bundled_axes = to_iter + list(method)
+        get_frame, after_bundle = _bundle(get_frame_dict[method], method,
+                                          to_iter, sizes, dtype)
+        return _transpose(get_frame, bundled_axes, result_axes)
+
+    # try to iterate without dropping axes
+    # sort in increasing number of dropped frames
+    # TODO: sometimes dropping some data is better than having many iterations
+    arr.sort(key=lambda x: x[5])
+    for method, axes_set, to_iter, n_iter, to_drop, n_drop in arr:
+        if n_iter > 0:
+            continue
+        get_frame, after_drop = _drop(get_frame_dict[method], method, to_drop)
+        return _transpose(get_frame, after_drop, result_axes)
+
+    # worst case: all methods have both too many axes and require iteration
+    # take lowest number of dropped frames
+    # if indecisive, take lowest number of iterations
+    arr.sort(key=lambda x: (x[3], x[5]))
+    method, axes_set, to_iter, n_iter, to_drop, n_drop = arr[0]
+
+    get_frame, after_drop = _drop(get_frame_dict[method], method, to_drop)
+    get_frame, after_bundle = _bundle(get_frame, after_drop, to_iter,
+                                      sizes, dtype)
+    return _transpose(get_frame, after_bundle, result_axes)
+
+
+class DefaultCoordsDict(dict):
+    """Dictionary that checks whether all keys are in `axes`"""
+    def __init__(self, default_coords=None):
+        """There is no check done here"""
+        super(DefaultCoordsDict, self).__init__()
+        self.axes = []
+
+    def __setitem__(self, attr, value):
+        if attr not in self.axes:
+            raise ValueError("axes %r does not exist" % attr)
+        super(DefaultCoordsDict, self).__setitem__(attr, value)
+
+    def update(self, *args, **kwargs):
+        # So that update does the check too
+        for k, v in six.iteritems(dict(*args, **kwargs)):
+            self[k] = v
+
 
 class FramesSequenceND(FramesSequence):
     """ A base class defining a FramesSequence with an arbitrary number of
@@ -299,12 +442,14 @@ class FramesSequenceND(FramesSequence):
     to which coordinates each index points. See below for a description of
     each attribute.
 
-    Subclassed readers only need to define `get_frame_2D`, `pixel_type` and
-    `__init__`. In the `__init__`, at least axes y and x need to be
-    initialized using `_init_axis(name, size)`.
+    Subclassed readers only need to define `pixel_type` and `__init__`. At least
+    one reader method needs to be registered as such using
+    `self._register_get_frame(method, <list of axes>)`.
+    In the `__init__`, axes need to be initialized using `_init_axis(name, size)`.
+    It is recommended to set default values to `bundle_axes` and `iter_axes`.
 
-    The attributes `__len__`, `frame_shape`, and `get_frame` are defined by
-    this base_class; these are not meant to be changed.
+    The attributes `__len__`, `get_frame`, and the attributes below are defined
+    by this base_class; these should not be changed by derived classes.
 
     Attributes
     ----------
@@ -318,26 +463,31 @@ class FramesSequenceND(FramesSequence):
         Shape of frames that will be returned by get_frame
     iter_axes : iterable of strings
         This determines which axes will be iterated over by the FramesSequence.
-        The last element in will iterate fastest. x and y are not allowed.
+        The last element in will iterate fastest. Default [].
     bundle_axes : iterable of strings
         This determines which axes will be bundled into one Frame. The axes in
         the ndarray that is returned by get_frame have the same order as the
-        order in this list. The last two elements have to be ['y', 'x'].
+        order in this list. Default ['y', 'x'].
     default_coords: dict of int
         When an axis is not present in both iter_axes and bundle_axes, the
-        coordinate contained in this dictionary will be used.
+        coordinate contained in this dictionary will be used. Default 0 for each.
 
     Examples
     --------
-    >>> class MDummy(FramesSequenceND):
+    >>> class DummyReaderND(FramesSequenceND):
     ...    @property
     ...    def pixel_type(self):
     ...        return 'uint8'
     ...    def __init__(self, shape, **axes):
+    ...        super(DummyReaderND, self).__init__()  # properly initialize
     ...        self._init_axis('y', shape[0])
     ...        self._init_axis('x', shape[1])
     ...        for name in axes:
     ...            self._init_axis(name, axes[name])
+    ...        self._register_get_frame(self.get_frame_2D, 'yx')
+    ...        self.bundle_axes = 'yx'  # set default value
+    ...        if 't' in axes:
+    ...            self.iter_axes = 't'  # set default value
     ...    def get_frame_2D(self, **ind):
     ...        return np.zeros((self.sizes['y'], self.sizes['x']),
     ...                        dtype=self.pixel_type)
@@ -348,21 +498,37 @@ class FramesSequenceND(FramesSequence):
     >>> frames.default_coords['m'] = 3
     >>> frames[5]  # returns Frame at T=5, M=3 with shape (2, 10, 64, 64)
     """
+    def __init__(self):
+        self._clear_axes()
+        self._get_frame_dict = dict()
+
+    def _register_get_frame(self, method, axes):
+        axes = tuple([a for a in axes])
+        if not hasattr(self, '_get_frame_dict'):
+            warn("Please call FramesSequenceND.__init__() at the start of the"
+                 "the reader initialization.")
+            self._get_frame_dict = dict()
+        self._get_frame_dict[axes] = method
+
     def _clear_axes(self):
         self._sizes = {}
-        self._default_coords = {}
+        self._default_coords = DefaultCoordsDict()
         self._iter_axes = []
         self._bundle_axes = ['y', 'x']
+        self._get_frame_wrapped = None
 
     def _init_axis(self, name, size, default=0):
         # check if the axes have been initialized, if not, do it here
         if not hasattr(self, '_sizes'):
+            warn("Please call FramesSequenceND.__init__() at the start of the"
+                 "the reader initialization.")
             self._clear_axes()
-        elif name in self._sizes:
+            self._get_frame_dict = dict()
+        if name in self._sizes:
             raise ValueError("axis '{}' already exists".format(name))
         self._sizes[name] = int(size)
-        if not (name == 'x' or name == 'y'):
-            self.default_coords[name] = int(default)
+        self.default_coords.axes = self.axes
+        self.default_coords[name] = int(default)
 
     def __len__(self):
         return int(np.prod([self._sizes[d] for d in self._iter_axes]))
@@ -392,76 +558,67 @@ class FramesSequenceND(FramesSequence):
         """ This determines which axes will be bundled into one Frame.
         The ndarray that is returned by get_frame has the same axis order
         as the order of `bundle_axes`.
-        The last two elements have to be ['y', 'x'].
         """
-        return self._bundle_axes
+        return self._bundle_axes[:]  # return a copy
 
     @bundle_axes.setter
     def bundle_axes(self, value):
+        value = list(value)
         invalid = [k for k in value if k not in self._sizes]
         if invalid:
             raise ValueError("axes %r do not exist" % invalid)
 
-        if len(value) < 2 or not (value[-1] == 'x' and value[-2] == 'y'):
-            raise ValueError("bundle_axes should end with ['y', 'x']")
-
         for k in value:
             if k in self._iter_axes:
                 del self._iter_axes[self._iter_axes.index(k)]
 
-        self._bundle_axes = list(value)
+        self._bundle_axes = value
+        if not hasattr(self, '_get_frame_dict'):
+            warn("Please call FramesSequenceND.__init__() at the start of the"
+                 "the reader initialization.")
+            self._get_frame_dict = dict()
+        if len(self._get_frame_dict) == 0:
+            if hasattr(self, 'get_frame_2D'):
+                # include get_frame_2D for backwards compatibility
+                self._register_get_frame(self.get_frame_2D, 'yx')
+            else:
+                raise RuntimeError('No reader methods found. Register a reader '
+                                   'method with _register_get_frame')
+
+        # update the get_frame method
+        get_frame = _make_get_frame(self._bundle_axes, self._get_frame_dict,
+                                    self.sizes, self.pixel_type)
+        self._get_frame_wrapped = get_frame
 
     @property
     def iter_axes(self):
         """ This determines which axes will be iterated over by the
-        FramesSequence. The last element will iterate fastest.
-        x and y are not allowed. """
-        return self._iter_axes
+        FramesSequence. The last element will iterate fastest. """
+        return self._iter_axes[:]  # return a copy
 
     @iter_axes.setter
     def iter_axes(self, value):
+        value = list(value)
         invalid = [k for k in value if k not in self._sizes]
         if invalid:
             raise ValueError("axes %r do not exist" % invalid)
 
-        if 'x' in value or 'y' in value:
-            raise ValueError("axes 'y' and 'x' cannot be iterated")
-
         for k in value:
             if k in self._bundle_axes:
                 del self._bundle_axes[self._bundle_axes.index(k)]
 
-        self._iter_axes = list(value)
+        self._iter_axes = value
 
     @property
     def default_coords(self):
         """ When a axis is not present in both iter_axes and bundle_axes, the
         coordinate contained in this dictionary will be used. """
-        return self._default_coords
+        return self._default_coords  # this is a custom dict (DefaultCoordsDict)
 
     @default_coords.setter
     def default_coords(self, value):
-        invalid = [k for k in value if k not in self._sizes]
-        if invalid:
-            raise ValueError("axes %r do not exist" % invalid)
-        if 'x' in value or 'y' in value:
-            raise ValueError("axes 'y' and 'x' cannot have a default "
-                             "coordinate")
         self._default_coords.update(**value)
 
-    @abstractmethod
-    def get_frame_2D(self, **ind):
-        """ The actual frame reader, defined by the subclassed reader.
-
-        This method should take exactly one keyword argument per axis,
-        reflecting the coordinate along each axis. It returns a two dimensional
-        ndarray with shape (sizes['y'], sizes['x']) and dtype `pixel_type`. It
-        may also return a Frame object, so that metadata will be propagated. It
-        will only propagate metadata if every bundled frame gives the same
-        fields.
-        """
-        pass
-
     def get_frame(self, i):
         """ Returns a Frame of shape determined by bundle_axes. The index value
         is interpreted according to the iter_axes property. Coordinates not
@@ -469,66 +626,29 @@ class FramesSequenceND(FramesSequence):
         value (see default_coords). """
         if i > len(self):
             raise IndexError('index out of range')
+        if self._get_frame_wrapped is None:
+            self.bundle_axes = tuple(self.bundle_axes)  # kick bundle_axes
 
         # start with the default coordinates
-        coords = self._default_coords.copy()
+        coords = self.default_coords.copy()
 
-        # list sizes of iterate axes
-        iter_sizes = [self._sizes[k] for k in self._iter_axes]
+        # list sizes of iteration axes
+        iter_sizes = [self._sizes[k] for k in self.iter_axes]
         # list how much i has to increase to get an increase of coordinate n
         iter_cumsizes = np.append(np.cumprod(iter_sizes[::-1])[-2::-1], 1)
         # calculate the coordinates and update the coords dictionary
         iter_coords = (i // iter_cumsizes) % iter_sizes
-        coords.update(**{k: v for k, v in zip(self._iter_axes, iter_coords)})
+        coords.update(**{k: v for k, v in zip(self.iter_axes, iter_coords)})
 
-        shape = self.frame_shape
-        if len(shape) == 2:  # simple case of only one frame
-            result = self.get_frame_2D(**coords)
-            if hasattr(result, 'metadata'):
-                metadata = result.metadata
-            else:
-                metadata = None
-        else:  # general case of N dimensional frame
-            Nframes = int(np.prod(shape[:-2]))
-            result = np.empty([Nframes] + list(shape[-2:]),
-                              dtype=self.pixel_type)
-
-            # zero out all coords that will be bundled
-            coords.update(**{k: 0 for k in self._bundle_axes[:-2]})
-            # read all 2D frames and properly iterate through the coordinates
-            mdlist = [{}] * Nframes
-            for n in range(Nframes):
-                frame = self.get_frame_2D(**coords)
-                result[n] = frame
-                if hasattr(frame, 'metadata'):
-                    mdlist[n] = frame.metadata
-                for dim in self._bundle_axes[-3::-1]:
-                    coords[dim] += 1
-                    if coords[dim] >= self._sizes[dim]:
-                        coords[dim] = 0
-                    else:
-                        break
-            # reshape the array into the desired shape
-            result.shape = shape
-
-            # propagate metadata
-            metadata = {}
-            if not np.all([md == {} for md in mdlist]):
-                keys = mdlist[0].keys()
-                for k in keys:
-                    try:
-                        metadata[k] = [row[k] for row in mdlist]
-                    except KeyError:
-                        # if a field is not present in every frame, ignore it
-                        warn('metadata field {} is not propagated')
-                    else:
-                        # if all values are equal, only return one value
-                        if metadata[k][1:] == metadata[k][:-1]:
-                            metadata[k] = metadata[k][0]
-                        else:  # cast into ndarray
-                            metadata[k] = np.array(metadata[k])
-                            metadata[k].shape = shape[:-2]
+        result = self._get_frame_wrapped(**coords)
+        if hasattr(result, 'metadata'):
+            metadata = result.metadata
+        else:
+            metadata = dict()
 
+        metadata_axes = set(self.axes) - set(self.bundle_axes)
+        metadata_coords = {ax: coords[ax] for ax in metadata_axes}
+        metadata.update(dict(axes=self.bundle_axes, coords=metadata_coords))
         return Frame(result, frame_no=i, metadata=metadata)
 
     def __repr__(self):
